#include <stdio.h>
#include <math.h>

#define USE_FLT

#include "common_macros.h"
#include "std_commonflt.h"
#include "geo_code.h"


typedef struct test_data test_data;


typedef test_data* divfuncdata;
#define DIVFUNCDATA divfuncdata


int SRC_TERM( DIVFUNCDATA fc,  TRI *sfrag, TRI *dfrag );


#define DIVFUNC subdivider
# define TERMFUNC SRC_TERM

#  include "divfunc.xmacro"

# undef TERMFUNC
#undef DIVFUNC




#ifndef FLOAT_TESTLEVEL
# define FLOAT_TESTLEVEL 1
#endif

#ifndef EQ_TESTLEVEL
# define EQ_TESTLEVEL 1
#endif

#ifndef CROSS_TESTLEVEL
# define CROSS_TESTLEVEL 1
#endif

#ifndef SPAR_TESTLEVEL
# define SPAR_TESTLEVEL 1
#endif

#ifndef TRIAREA_TESTLEVEL
# define TRIAREA_TESTLEVEL 1
#endif

#ifndef VSINTER_TESTLEVEL
# define VSINTER_TESTLEVEL 1
#endif

#ifndef DEBUG_DEPTHBRAKE
# define DEBUG_DEPTHBRAKE 0
#endif



void eq_test( FLT_TYPE a, FLT_TYPE b, FLT_TYPE eps )
{
	printf
	(
		"( %g, %g, %g )\n    { EQ = %i, GR = %i, LS = %i, GREQ = %i, LSEQ = %i };\n",
		a,
		b,
		eps,
		
		(int)( EQ( a, b, eps ) ),
		(int)( GR( a, b, eps ) ),
		(int)( LS( a, b, eps ) ),
		(int)( GREQ( a, b, eps ) ),
		(int)( LSEQ( a, b, eps ) )
	);
}
void print_tri( const char *preq, TRI *t )
{
	PRINT_TRI( preq, t, 1 );
}
void print_vecpar( const char *preq, TRI *t,  int a, int b )
{
	if( t )
	{
		if( !preq )
		{
			preq = "";
		}
		
		int res = VEC_PAR
			(
				t->x[ a ], t->y[ a ],
				t->x[ b ], t->y[ b ]
			);
		printf
		(
			"%sVEC_PAR( %f, %f,  %f, %f ) : %i;\n",
			preq,
			
			t->x[ a ], t->y[ a ],
			t->x[ b ], t->y[ b ],
			
			res
		);
	}
}



struct test_data
{
	TRI t;
	VS vert_set;
};


int main( int argn, char *argc[] )
{
	test_data fc;
	TRI t_clip;
	FLT_TYPE cut1x[ 8 ] =
				{ 1, 2, 3, 4, 5, 6, 7, 8 },
			 cut2x[ 8 ] =
				{ 1, 2, 3, 4, 5, 6, 7, 8 };
	FLT_TYPE cut1y[ 8 ] =
				{ 4, 4, 4, 4, 4, 4, 4, 4 },
			 cut2y[ 8 ] =
				{ -4, -4, -4, -4, -4, -4, -4, -4 };
	size_t i = 0;
	int ires, reti[ 3 ];
	FLT_TYPE retf[ 3 ];
	float ftest;
	
	
	
	printf( "\n\n" );
	
	
	printf( "EPSILON = %e;\n-EPSILON = %e;\n\n", (float)( EPSILON ), -(float)( EPSILON ) );
	
	
	printf( "FLOAT_TESTLEVEL == %i;\n", (int)( FLOAT_TESTLEVEL ) );
	printf( "EQ_TESTLEVEL == %i;\n", (int)( EQ_TESTLEVEL ) );
	printf( "CROSS_TESTLEVEL == %i;\n", (int)( CROSS_TESTLEVEL ) );
	printf( "SPAR_TESTLEVEL == %i;\n", (int)( SPAR_TESTLEVEL ) );
	printf( "TRIAREA_TESTLEVEL == %i;\n", (int)( TRIAREA_TESTLEVEL ) );
	printf( "VSINTER_TESTLEVEL == %i;\n", (int)( VSINTER_TESTLEVEL ) );
	printf( "DEBUG_DEPTHBRAKE == %i;\n", (int)( DEBUG_DEPTHBRAKE ) );
	
	printf( "\n\n" );
	
	
	
#if FLOAT_TESTLEVEL > 0
	printf( "Floating-point test:\n" );
	
	ftest = 0.5;
	printf( "0.5: %f;\n", (float)( ftest ) );
	
	ftest = 1;
	printf( "1 / 2: %f;\n", (float)( ftest / 2 ) );
	
	ftest = 1;
	printf( "abs( 1 ): %f;\n", (float)abs( ftest ) );
	
	ftest = 1;
	printf( "abs( 1 ) / 2: %f;\n", (float)( abs( ftest ) / 2 ) );
	
	ftest = 1;
	printf( "abs( 1 ) / 2.0: %f;\n", (float)( abs( ftest ) / 2.0 ) );
	
	ftest = 1;
	printf( "abs( 1 ) * 0.5: %f;\n", (float)( abs( ftest ) * 0.5 ) );
#endif
	
	
	
	/* Set fc.t & fc.t_clip */
	
	fc.t.x[ 0 ] = 0;
	fc.t.y[ 0 ] = 0;
	
	fc.t.x[ 1 ] = 2;
	fc.t.y[ 1 ] = 0;
	
	fc.t.x[ 2 ] = 0;
	fc.t.y[ 2 ] = 2;
	
	
	fc.vert_set.x1 = cut1x;
	fc.vert_set.y1 = cut1y;
	
	fc.vert_set.x2 = cut2x;
	fc.vert_set.y2 = cut2y;
	
	fc.vert_set.count = sizeof( cut1x ) / sizeof( FLT_TYPE );
	
	
	/* This enables complete shut-down of this particular set of tests. */
#if EQ_TESTLEVEL > 0
	eq_test( 0.0, 0.0, EPSILON );
	
	printf( "\n" );
	
	eq_test( 0.0, 0.0, 1.0 );
	
	printf( "\n" );
	
	eq_test( 1.0, 0.0, 1.0 );
	eq_test( 0.0, 1.0, 1.0 );
	
#if EQ_TESTLEVEL > 1
	printf( "\n" );
	
	eq_test( 1.0, 0.0, 2.0 );
	eq_test( 0.0, 1.0, 2.0 );
#endif
	
	printf( "\n\n\n" );
	
	eq_test( EPSILON, 0.0, EPSILON );
	eq_test( 0.0, EPSILON, EPSILON );
	
#if EQ_TESTLEVEL > 1
	printf( "\n" );
	
	eq_test( -EPSILON, 0.0, EPSILON );
	eq_test( 0.0, -EPSILON, EPSILON );
#endif
	
	printf( "\n\n" );
	
	eq_test( EPSILON, 0.0, -EPSILON );
#if EQ_TESTLEVEL > 1
	eq_test( 0.0, EPSILON, -EPSILON );
	
	printf( "\n" );
	
	eq_test( -EPSILON, 0.0, -EPSILON );
	eq_test( 0.0, -EPSILON, -EPSILON );
#endif
	
	printf( "\n\n\n" );
	
	eq_test( EPSILON, -EPSILON, EPSILON );
	eq_test( EPSILON, -EPSILON, EPSILON * 2 );
	
#if EQ_TESTLEVEL > 1
	printf( "\n" );
	
	eq_test( EPSILON, -EPSILON, -EPSILON );
	eq_test( EPSILON, -EPSILON, -EPSILON * 2 );
	
	printf( "\n\n" );
	
	eq_test( -EPSILON, EPSILON, EPSILON );
	eq_test( -EPSILON, EPSILON, EPSILON * 2 );
	
	printf( "\n" );
	
	eq_test( -EPSILON, EPSILON, EPSILON );
	eq_test( -EPSILON, EPSILON, EPSILON * 2 );
#endif
	
	printf( "\n\n\n" );
	
	eq_test( 8 - 4, 2 * 2, EPSILON );
	
	eq_test( 1, 1, EPSILON );
	
	printf( "\n" );
	
	eq_test( 1, 2, EPSILON );
	eq_test( 2, 1, EPSILON );
	
	printf( "\n\n" );
#endif
	
	
	
	print_tri( "test ", &( fc.t ) );
	
	
	printf( "\n\n" );
	
	
#if CROSS_TESTLEVEL > 0
	printf
	(
		"v1.x * v1.y - v1.x * v1.y = %e;\n  abs() = %e;\n",
		(float)( fc.t.x[ 0 ] * fc.t.y[ 0 ] - fc.t.x[ 0 ] * fc.t.y[ 0 ] ),
		(float)abs( fc.t.x[ 0 ] * fc.t.y[ 0 ] - fc.t.x[ 0 ] * fc.t.y[ 0 ] )
	);
	printf
	(
		"abs( v1.x * v1.y - v1.x * v1.y ) : %f;\n",
		abs( fc.t.x[ 0 ] * fc.t.y[ 0 ] - fc.t.x[ 0 ] * fc.t.y[ 0 ] )
	);
	eq_test( abs( fc.t.x[ 0 ] * fc.t.y[ 0 ] - fc.t.x[ 0 ] * fc.t.y[ 0 ] ), 0.0, EPSILON );
	
	printf( "\n" );
	
	printf
	(
		"v1.x * v2.y - v2.x * v1.y = %e;\n  abs() = %e;\n",
		(float)( fc.t.x[ 0 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 0 ] ),
		(float)abs( fc.t.x[ 0 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 0 ] )
	);
	
	printf
	(
		"abs( v1.x * v2.y - v2.x * v1.y ) : %f;\n",
		abs( fc.t.x[ 0 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 0 ] )
	);
	eq_test( abs( fc.t.x[ 0 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 0 ] ), 0.0, EPSILON );
	
	printf( "\n" );
	
	printf
	(
		"v3.x * v2.y - v2.x * v3.y = %e;\n  abs() = %e;\n",
		(float)( fc.t.x[ 2 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 2 ] ),
		(float)abs( fc.t.x[ 2 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 2 ] )
	);
	printf
	(
		"abs( v3.x * v2.y - v2.x * v3.y ) : %f;\n",
		abs( fc.t.x[ 2 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 2 ] )
	);
	eq_test( abs( fc.t.x[ 2 ] * fc.t.y[ 1 ] - fc.t.x[ 1 ] * fc.t.y[ 2 ] ), 0.0, EPSILON );
	
	
	printf( "\n\n" );
#endif
	
	
#if SPAR_TESTLEVEL > 0
	print_vecpar( "", &( fc.t ),  0, 0 );
	print_vecpar( "", &( fc.t ),  0, 1 );
	
	print_vecpar( "", &( fc.t ),  1, 1 );
	
	print_vecpar( "", &( fc.t ),  1, 2 );
	
	
	printf( "\n\n" );
#endif
	
	
	
#if TRIAREA_TESTLEVEL > 0
	/* Area testing. */
	
	fc.t.x[ 0 ] = 0;
	fc.t.y[ 0 ] = 0;
	
	fc.t.x[ 1 ] = 0;
	fc.t.y[ 1 ] = 0;
	
	fc.t.x[ 2 ] = 0;
	fc.t.y[ 2 ] = 0;
	
	
	
	printf( "\nArea test:\n" );
	
	fc.t.x[ 1 ] = 0;
	fc.t.y[ 2 ] = 0;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	printf( "\n" );
	
	fc.t.x[ 1 ] = 0;
	fc.t.y[ 2 ] = 1;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 1;
	fc.t.y[ 2 ] = 0;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 1;
	fc.t.y[ 2 ] = 1;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	printf( "\n" );
	
	fc.t.x[ 1 ] = 1;
	fc.t.y[ 2 ] = 2;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 2;
	fc.t.y[ 2 ] = 1;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 2;
	fc.t.y[ 2 ] = 2;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	printf( "\n" );
	
	fc.t.x[ 1 ] = 1;
	fc.t.y[ 2 ] = 3;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 3;
	fc.t.y[ 2 ] = 1;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 2;
	fc.t.y[ 2 ] = 3;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 3;
	fc.t.y[ 2 ] = 2;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	fc.t.x[ 1 ] = 3;
	fc.t.y[ 2 ] = 3;
	print_tri( "", &( fc.t ) );
	printf( "area: %f;\n\n", (float)( TRI_AREA_INLINE( &( fc.t ), (int*)0 ) ) );
	
	printf( "\n\n" );
	
	
	
	/* Restore the previous triangle values. */
	
	fc.t.x[ 0 ] = 0;
	fc.t.y[ 0 ] = 0;
	
	fc.t.x[ 1 ] = 2;
	fc.t.y[ 1 ] = 0;
	
	fc.t.x[ 2 ] = 0;
	fc.t.y[ 2 ] = 2;
#endif
	
	
#if VSINTER_TESTLEVEL > 0
	printf( "vs count: %i;\n", (int)fc.vert_set.count );
	i = 0;
	while( i < fc.vert_set.count )
	{
		printf
		(
			"  vs ( x1 = %f, y1 = %f ), ( x2 = %f, y2 = %f ) [ %i ];\n",
			fc.vert_set.x1[ i ], fc.vert_set.y1[ i ],
			fc.vert_set.x2[ i ], fc.vert_set.y2[ i ],
			(int)i
		);
		
		ires = FIND_INTERS ( &( fc.t ),  fc.vert_set.x1[ i ], fc.vert_set.y1[ i ],  fc.vert_set.x2[ i ], fc.vert_set.y2[ i ],    reti, retf );
		printf( "  FIND_INTERS( &test_tri, x1, y1,  x2, y2,  &reti, &retf) : %i;\n", ires );
		printf( "   reti = { %i, %i, %i };\n",  reti[ 0 ], reti[ 1 ], reti[ 2 ] ),
		printf( "   retf = { %f, %f, %f };\n",  retf[ 0 ], retf[ 1 ], retf[ 2 ] );
		
		printf( "\n" );
		++i;
	}
#endif
	
	
	
	{
		FLT_TYPE bx, by;
		FLT_TYPE rx, ry;
		
		
		print_tri( "bary reference tri ", &( fc.t ) );
		
		
		rx = 1;
		ry = 1;
		printf( "  input position: ( %f, %f );\n",  (float)rx, (float)ry );
		FIND_BARY( &( fc.t ),  rx, ry,    &bx, &by );
		printf( "  barycentric translation: ( %f, %f );\n",  (float)bx, (float)by );
		USE_BARY( &( fc.t ),  bx, by,    &rx, &ry );
		printf( "  recovered position: ( %f, %f );\n",  (float)rx, (float)ry );
	}
	
	
	
	print_tri( "input ", &( fc.t ) );
	printf( "\n" );
	
	
	subdivider( &fc,  &( fc.t ), &( fc.t ), &( fc.vert_set ) );
	
	
	/* We've done vertical cuts, time for horizontal. */
	
	fc.vert_set.x1 = cut1y;
	fc.vert_set.y1 = cut1x;
	
	fc.vert_set.x2 = cut2y;
	fc.vert_set.y2 = cut2x;
	
	subdivider( &fc,  &( fc.t ), &( fc.t ), &( fc.vert_set ) );
}


int SRC_TERM( DIVFUNCDATA fc,  TRI *sfrag, TRI *dfrag )
{
	printf( "SRC_TERM entered.\n" );
	
	print_tri( "SRC_TERM sfrag ", sfrag );
	print_tri( "SRC_TERM dfrag ", dfrag );
	
	printf
	(
		"{ 0, 1, 2 } PT_IN_TRI( dfrag, sfrag[] ) = { %i, %i, %i };\n",
		
		(int)PT_IN_TRI ( dfrag,  sfrag->x[ 0 ], sfrag->y[ 0 ] ),
		(int)PT_IN_TRI ( dfrag,  sfrag->x[ 1 ], sfrag->y[ 1 ] ),
		(int)PT_IN_TRI ( dfrag,  sfrag->x[ 2 ], sfrag->y[ 2 ] )
	);
	
	printf( "\n\n" );
	
	return( 1 );
}
